/*  pcmplay v1.0.1  Copyright (C) 1996  Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* pcmplay.c */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#define GRDL_VER "v1.0.1"


/* reads len number of bytes from file and turns them into integer format */
int getref(f, len)
FILE *f;
int len;
{
  unsigned int a=0, b;

  for(b=0;b < len;b++)
    a=(a<<8)+getc(f);
  return a;
}

void main(argc, argv)
int argc;
char *argv[];
{
  FILE *f;
  char buf[48126], *cmd=argv[0];
  int a, b, c, vol, typ, loop;
  int count, len, event, play, relay, speed=16000, timer, maxtime;
  int opt=0, loopon=0, backgroundonly=0, continuous=0;
  int audio, mixer;

  /* grab any flags on the command line */
  while(argc > 1) {
    --argc; ++argv;
    if(opt) {
      if(!(speed=atoi(*argv)))
        speed=16000;
      opt=0;
      continue;
    }
    if(**argv == '-') {
      for(a=1;(*argv)[a];a++) {
        switch((*argv)[a]) {
        case 'b':
          backgroundonly=1;
          break;
        case 'c':
          continuous=1;
          break;
        case 'r':
          opt|=1;
          break;
        case 'v':
          fprintf(stdout, "%s: GRDL Version Reference %s\n", cmd, GRDL_VER);
          fflush(stdout);
          exit(0);
        default:
          fprintf(stderr, "%s: illegal option -- %c\n", cmd, (*argv)[a]);
          fprintf(stderr, "Usage: %s [-bcv] [-r rate] file ...\n", cmd);
          fflush(stderr);
          exit(1);
        }
      }
      continue;
    } else
      break;
  }
  /* try to open the audio and mixer files */
  if((audio = open("/dev/dsp", O_WRONLY, 0)) < 0) {
    perror("/dev/dsp");
    exit(1);
  } if((mixer = open("/dev/mixer", O_RDWR, 0)) < 0) {
    perror("/dev/mixer");
    exit(1);
  }

  /* loop through the files on commandline */
  for(;argc;argc--,++argv) {
    if(!(f=fopen(*argv, "r")))
      continue;
    printf("\e[2J\e[H\nNow Playing:%50sLunar PCM Player\n", "");
    printf("%60sby Byron Stanoszek\n", "");
    printf("   %-55.55s   gandalf@winds.org\n\n", *argv);

    /* set startup volume: 75% L, 75% R */
    a=(75*256)+75;
    ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a); /* output volume */

    /* sound buffer: 2 fragments of 4096 bytes each */
    a=(2<<16)+12;
    ioctl(audio, SNDCTL_DSP_SETFRAGMENT, &a);

    /* begin by reading file header */

    typ=getref(f, 2);
    loop=getref(f, 4);
    len=getref(f, 4);

    if(len-(loop*1024) < 16384)
      loopon=0;
    else
      loopon=1;
    event=fgetc(f);

    printf("File Type: %s\n", typ==0?"Lunar I: The Silver Star PCM":
          (typ==2)?(event?"Lunar II: Eternal Blue - Background Music":
          (!loop?"Null Sound File":"Lunar II: Eternal Blue - Event Dialogue")):
          "Unknown Format");

    if(!event && !loop) {
      loop=-1;
      vol=0;
    } else
      vol=75;

    if(typ) {
      printf("Samples: %d\n", len);
      if(loop >= 0 && loopon)
        printf("Loop Point: %d Block%s\n\n", loop, loop==1?"":"s");
    }

    a=0; ioctl(audio, SNDCTL_DSP_STEREO, &a);
    a=8; ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &a);
    ioctl(audio, SNDCTL_DSP_SPEED, &speed);

    printf("Sample Rate: %d Hz\n\n\n\n", speed);

    /* Determine length of tune */
    if(continuous)
      maxtime=359999;
    else if(loopon)
      maxtime=((len+(len-(loop*1024)))/speed)+16;
    else
      maxtime=len/speed;


    /* Play Lunar I Format */
    if(!typ) {
      b=0;
      while((c=fgetc(f)) != EOF) {
        if(c > 127)
          c=127-c;
        else
          c--;
        buf[b++]=c+128;
        if(b == speed) {
          write(audio, buf, b);
          b=0;
        }
      }
      write(audio, buf, b);
      continue;
    }

    /* Play Lunar II Format */

    /* start of music */
    fseek(f, 2048, SEEK_SET);

    count=0; b=0; play=0; relay=64; timer=0;

    if(typ != 2) /* unknown format */
      vol=0;
    if(backgroundonly && (!event || loop == -1))
      vol=0;

    while((c=fgetc(f)) != EOF && vol) {
      if(typ)
        c=fgetc(f);
      if(c > 127)
        c=127-c;
      else
        c--;
      buf[b++]=c+128;
      if(b == speed) {  /* each second.. */
        if(timer > -1 && maxtime-timer >= 0) {
          if(continuous)
            printf("\e[ATime Playing: %02d:%02d:%02d\n", timer/3600,
                   (timer%3600)/60, timer%60);
          else
            printf("\e[ATime Playing: %02d:%02d        Time Remaining: %02d:%02d\n",
                   timer/60, timer%60, (maxtime-timer)/60, (maxtime-timer)%60);
        }
        if((!loopon || !typ) && maxtime-timer < 1)
          break;
        write(audio, buf, b);
        b=0;
        ++timer;
      }
      if(!(b%(speed/8)) && count > 1 && !continuous && --relay < 1) {
        vol--;
        a=(vol*256)+vol;
        ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a);
      }
      if(++play == len) {
        fseek(f, (loop+1)*2048, SEEK_SET);
        play=loop*1024;
        count++;
      }
    }

    /* clear audio before continuing */
    ioctl(audio, SNDCTL_DSP_RESET);
    fclose(f);
  }

  /* clean up all files and exit */

  close(mixer);
  close(audio);

  printf("\n\n");
  fflush(stdout);
  exit(0);
}
